//
//  C_Helper.c
//  Part of the Xojo project "zlib compression.rbp"
//
//  Created by Thomas Tempelmann on 30 Jun 2016.
//
// On macOS, build with:
//   gcc -c C_Helper.c -DTARGET_OS_MAC -arch i386 -arch x86_64
//   gcc -dynamiclib -undefined suppress -flat_namespace C_Helper.o -o C_Helper.dylib -arch i386 -arch x86_64


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <zlib.h>
#include <string.h>
#include <assert.h>

/*
 * Note:
 *
 * Functions declared as "extern" and starting with "helper_" are meant to be accessed by the Xojo code.
 */

#if TARGET_OS_MAC
	#include <CoreFoundation/CoreFoundation.h>
	void NSLog(CFStringRef format, ...);
	#define LOGF(s, ...) NSLog(CFSTR(s), __VA_ARGS__)
	#define DLL
#else
	// Windows
	typedef signed char int8_t;
	typedef unsigned char uint8_t;
	typedef unsigned short uint16_t;
	typedef long int32_t;
	typedef unsigned long uint32_t;
	typedef unsigned long long uint64_t;
	#define LOGF(s, ...) (void*)0
	#define DLL __declspec(dllexport) __stdcall
#endif

typedef int boolean;
#define false 0
#define true 1

static int gLog = 0;	// Bits: 0: log issues, 1: log everything
static pthread_t *gThreads = NULL;
static int gThreadCount = 0;
static int gStop = 0;

typedef struct {
	uint16_t ident;				// must equal 0xB117
	uint16_t ownSize;			// must equal sizeof(privateData)
	struct sharedTask *next;
} privateData;

typedef struct sharedTask {
	uint16_t ident;				// must equal 0xB119
	uint16_t ownSize;			// must equal sizeof(sharedTask)
	privateData internal;
	uint8_t status;				// 0: ready, 1: compressing, 2: finished, 3: error
	uint8_t compressionLevel;	// 1 (best speed) to 9 (best compression)
	int8_t  wbits;
	uint8_t finish;
	uint32_t userInfo;
	uint32_t inputLen;
	uint32_t outputBufSize;
	uint32_t outputLen;
	void *inputPtr;
	void *outputPtr;
} sharedTask;

static sharedTask *gNextTask = NULL;
static sharedTask *gLastTask = NULL;
static int gPendingTaskCount = 0;
static int gBusyTaskCount = 0;

static pthread_mutex_t gInputLock;
static pthread_cond_t gInputCond = PTHREAD_COND_INITIALIZER;


/*
 * Compression in thread
 */

static void *threadRunner(void *threadNum)
{
	if (gLog & 2) LOGF("Thread %d ready", (int)threadNum);
	
	/*
	 * This is a typical Worker thread:
	 * It puts itself to sleep and waits to be woken up.
	 * Once it's awake, it'll check if there's work to do.
	 * If so, it performs the task and then loops back.
	 */

	int level = -2;
	int wbits = 0;
	z_stream zs = {0};
	int res;
	
	while (!gStop) {
		sharedTask *myTask;

		pthread_mutex_lock (&gInputLock);
		// wait for task
		while (gNextTask == NULL) {
			pthread_cond_wait (&gInputCond, &gInputLock);
			if (gStop) {
				pthread_mutex_unlock (&gInputLock);
				goto bail;
			}
		}
		// pick next task from queue
		myTask = gNextTask;
		gNextTask = myTask->internal.next;
		gPendingTaskCount -= 1;
		pthread_mutex_unlock (&gInputLock);

		myTask->status = 1;
		gBusyTaskCount += 1;
		if (gLog & 2) LOGF("#%d compressing on thread %d...", (int)myTask->userInfo, (int)threadNum);

		if (level != myTask->compressionLevel || wbits != myTask->wbits) {
			// set compression level and wbits
			level = myTask->compressionLevel;
			wbits = myTask->wbits;
			res = deflateInit2 (&zs, level, Z_DEFLATED, wbits, 8, Z_DEFAULT_STRATEGY);
			if (res && gLog) LOGF("#%d deflateInit() err %d", (int)myTask->userInfo, res);
		} else {
			res = deflateReset (&zs);
			if (res && gLog) LOGF("#%d deflateReset() err %d", (int)myTask->userInfo, res);
		}
		
		zs.next_in = myTask->inputPtr;
		zs.avail_in = myTask->inputLen;
		zs.next_out = myTask->outputPtr;
		zs.avail_out = myTask->outputBufSize;
		zs.total_out = 0;
		
		int mode = myTask->finish ? Z_FINISH : Z_SYNC_FLUSH;
		res = deflate (&zs, mode);	// finish or continue
		if (zs.avail_out == 0) {
			if (gLog) LOGF("#%d deflate(): avail_out == 0", (int)myTask->userInfo);
			res = -200;
		}
		
		if (res == Z_STREAM_END) {
			if (gLog & 2) LOGF("#%d deflate() stream end", (int)myTask->userInfo);
			res = 0;
		} else if (res == Z_OK) {
			if (gLog & 2) LOGF("#%d deflate() ok", (int)myTask->userInfo);
		} else {
			if (gLog) LOGF("#%d deflate() err %d", (int)myTask->userInfo, res);
		}
		
		myTask->outputLen = zs.total_out;

		// Mark the task finished
		myTask->status = (res == 0) ? 2 : 3;
		gBusyTaskCount -= 1;
	}

bail:
	// free the zlib stream buffer
	res = deflateEnd (&zs);
	if (res && gLog) {
		// We can usually ignore this error as it always gives us Z_DATA_ERROR
		if (gLog & 2 || res != Z_DATA_ERROR) LOGF("deflateEnd() err %d on thread %d", res, (int)threadNum);
	}
	
	// wake other threads, so they check the gStop condition as well
	pthread_cond_signal (&gInputCond);

	if (gLog & 2) LOGF("Thread %d finished", (int)threadNum);

	return NULL;
}

DLL extern boolean helper_enqueue (sharedTask *task)
{
	if (task->ident != 0xB119 || task->internal.ident != 0xB117) return false;
	
	if (task->ownSize != sizeof(*task)) {
		LOGF("Error! ownSize: %d, sizeof: %d", (int)task->ownSize, (int)sizeof(*task));
		return false;
	}
	
	if (task->internal.ownSize != sizeof(task->internal)) {
		LOGF("Error! internal ownSize: %d, sizeof: %d", (int)task->internal.ownSize, (int)sizeof(task->internal));
		return false;
	}

	task->internal.next = NULL;

	// Add the task to the queue.
	pthread_mutex_lock(&gInputLock);
	if (!gNextTask) {
		// The queue was empty - we start fresh
		gNextTask = task;
	} else {
		// Append this task to the end of the queue
		gLastTask->internal.next = task;
	}
	gLastTask = task;
	gPendingTaskCount += 1;
	pthread_mutex_unlock(&gInputLock);

	// Signal the waiting threads so that one picks up the task
	pthread_cond_signal (&gInputCond);

	return true;
}

DLL extern boolean helper_disposeThreads ()
{
	// Stop threads from processing more tasks
	pthread_mutex_lock (&gInputLock);
	gNextTask = NULL;
	gLastTask = NULL;
	gStop = 1;
	pthread_mutex_unlock (&gInputLock);

	if (gThreads) {
		// let's wake our threads so that they see the gStop flag and quit
		pthread_cond_signal (&gInputCond);

		// Wait for all threads to end before we free their memory
		for (int n = 0; n < gThreadCount; ++n) {
			pthread_t *th = &gThreads[n];
			pthread_join (*th, NULL);
		}

		free (gThreads);
	}
	gThreads = NULL;
	gThreadCount = 0;

	gPendingTaskCount = 0;
	gBusyTaskCount = 0;

	return true;
}

DLL extern boolean helper_createThreads (int threadCountIn)
{
	// The memory for the "pthread_t" elements needs to be static,
	// hence we allocate it now dynamically

	if (gThreads) {
		// Error: There's already one allocated
		return false;
	}

	gStop = 0;
	gThreads = calloc (threadCountIn, sizeof(gThreads[0]));
	gThreadCount = threadCountIn;

	for (int n = 0; n < gThreadCount; ++n) {
		pthread_create (&gThreads[n], NULL, threadRunner, (void*)(size_t)n);
	}

	return true;
}

DLL extern boolean helper_getCounts (int32_t *threadsOut, int32_t *pendingOut, int32_t *busyOut)
{
	*threadsOut = gThreadCount;
	*pendingOut = gPendingTaskCount;
	*busyOut = gBusyTaskCount;
	return true;
}

DLL extern boolean helper_initialize (int logging)
{
	gLog = logging;
	pthread_mutex_init (&gInputLock, NULL);
	return true;
}
